/****************************************************************************
 * tgdisasm.c -- Joseph LoCicero, IV -- 96/5/09
                 Dave Shadoff        -- 96/5/09

    Simple TG Disassembler based on 65C02 opcode listing.

    Updates:
    --------
    Dave Shadoff  96/7/19  Place 'new' Hu6280 opcodes/addressing modes in,
                           due to new information available from Japanese
                           'Develo' system asembler.

 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <math.h>
#include <sys\stat.h>
#include <errno.h>

#include "format.h"
#include "dis.h"


/* DEFINE the line below if running under OS/2 (no ESPIPE) */
#define  USING_OS2


/* buffer size of opcode buffer */

#define  OPBUF_SIZE    7


/* GLOBALS */

char opbuf[OPBUF_SIZE];



/* jologetchar, just to keep code clean, does a nice read(). */
int jologetchar(FILE *inf)
{
   unsigned char c;

   return ((fread(&c, 1, 1, inf) == 0) ? EOF: (int) c);
}


long inputnum(char * string)
{
   long   value = 0;
   char * c = (string + 1);

   if (*string != '$') {
      return(atol(string));
   }

   while (*c != '\0') {

     value *= 16;

     if ((*c >= '0') && (*c <= '9')) {
       value += (*c - '0');

     } else if ((*c >= 'A') && (*c <= 'F')) {
       value += (*c - 'A' + 10);

     } else if ((*c >= 'a') && (*c <= 'f')) {
       value += (*c - 'a' + 10);

     } else {

       return(-1);
     }

     
     c++;
   }

   return(value);
}


int main(int argc, char *argv[])
{
   FILE *inf;
   FILE *outf;
   char linebuf[256];
   unsigned char op;
   long ctr    = 0;
   long offset = 0;
   long length = -1;
   int  i, size;
   int  temp;

   if (argc < 3) {
      fprintf(stderr, "Usage:  %s <infile> <outfile> [length] [offset] [startaddr]\n", argv[0]);
      fprintf(stderr, "   [length]     number of instructions to parse\n");
      fprintf(stderr, "   [offset]     first byte of the input file to parse\n");
      fprintf(stderr, "   [startaddr]  outfile's starting address\n");
      fprintf(stderr, " (numbers default to decimal; hexadecimal if preceeded by '$')\n");
      exit(1);
   } else {
      /* Open input and output files. */
      if ((inf = fopen(argv[1], "rb")) == NULL) {
         fprintf(stderr, "%s: can't open %s\n", argv[0], argv[1]);
         exit(1);
      }
      if ((outf = fopen(argv[2], "w")) == NULL) {
         fprintf(stderr, "%s: can't open %s\n", argv[0], argv[2]);
         exit(1);
      }

      /* length if requested */
      if (argc >= 4) {
         length = inputnum(argv[3]);
         if (length == -1) {
            fprintf(stderr, "Invalid length!");
            exit(1);
         }
      }

      /* Skip to a certain offset if requested */
      if (argc >= 5) {
         offset = inputnum(argv[4]);
         if (offset == -1) {
            fprintf(stderr, "Invalid offset!");
            exit(1);
         }
         if (fseek(inf, offset, SEEK_SET)) {
            fprintf(stderr, "End of input file reached!");
            exit(1);
         }
#ifdef USING_OS2
         if (errno == EBADF || errno == EINVAL) {
#else
         if (errno == EBADF || errno == EINVAL || errno == ESPIPE) {
#endif
            fprintf(stderr, "End of input file reached!");
            exit(1);
         }
      }

      /* Start numbering at requested integer */
      if (argc == 6) {
         ctr = inputnum(argv[5]);
         if (ctr == -1) {
            fprintf(stderr, "Invalid start address !");
            exit(1);
         }
      }

      for (i = 0; i < OPBUF_SIZE; i++) {
         if ((temp = jologetchar(inf)) == EOF) {
            fprintf(stderr, "early file end !\n");
            exit(1);
         }
         opbuf[i] = temp;
      }

      while (length != 0) {

         op = opbuf[0];

         /* This line is the real 'meat' of the disassembler: */

         (*addr_info[optable[op].addr_mode].func)           /* function      */
              (linebuf, ctr, opbuf, optable[op].opname);    /* parm's passed */

         fprintf(outf, "%s\n", linebuf);

         length--;
         if (!length)
            break;

         size = addr_info[optable[op].addr_mode].size;
         ctr += size;

         /* move buffer contents back */
         for (i = size; i < OPBUF_SIZE; i++) {
            opbuf[i-size] = opbuf[i];
         }

         /* refill buffer */
         for (i = OPBUF_SIZE-size; i < OPBUF_SIZE; i++) {
            if ((temp = jologetchar(inf)) == EOF) {
               fprintf(stderr, "early file end !\n");
               exit(1);
            }
            opbuf[i] = temp;
         }
      } /* End of while read loop */
   } /* endif Were there enough arguments? */
   return(0);                           /* return value to appease compiler */
}
